/* -*- mode: C; mode: fold; -*- */
/*
Copyright (C) 2011 John E. Davis

This file is part of the S-Lang Voronoi Module

The S-Lang Voronoi Module is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License as
published by the Free Software Foundation; either version 2 of the
License, or (at your option) any later version.

The S-Lang Voronoi Module is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.  
*/

#include "config.h"

#include <stdio.h>
#include <slang.h>
#ifdef HAVE_STDLIB_H
# include <stdlib.h>
#endif
#include <string.h>
#include <math.h>

#include "voronoi.h"

#ifdef __cplusplus
extern "C"
{
#endif
SLANG_MODULE(voronoi);
#ifdef __cplusplus
}
#endif

#include "version.h"

typedef struct
{
   double xmin, xmax, ymin, ymax;
}
Bound_Type;

typedef struct _Node_Type
{
   int type;
#define NODE_TYPE_NODE		1
#define NODE_TYPE_BNDS_NODE	2
#define NODE_TYPE_LEAF		3
   double median;
   struct _Node_Type *left;
   struct _Node_Type *right;
}
Node_Type;


typedef struct
{
   int type;
   double median;
   struct _Node_Type *left;
   struct _Node_Type *right;
   Bound_Type bnds;
}
Bound_Node_Type;


typedef struct _Leaf_Node_Type
{
   int type;
   Bound_Type bnds;
   SLindex_Type *indices;
   SLuindex_Type num_indices;
}
Leaf_Node_Type;

typedef struct Node_Chunk_Type
{
   struct Node_Chunk_Type *next;
#define NODE_CHUNK_SIZE 256
   unsigned int num_taken;
   Node_Type nodes[NODE_CHUNK_SIZE];
}
Node_Chunk_Type;

typedef struct Leaf_Node_Chunk_Type
{
   struct Leaf_Node_Chunk_Type *next;
#define LEAF_NODE_CHUNK_SIZE 256
   unsigned int num_taken;
   Leaf_Node_Type nodes[LEAF_NODE_CHUNK_SIZE];
}
Leaf_Node_Chunk_Type;

typedef struct Bound_Node_Chunk_Type
{
   struct Bound_Node_Chunk_Type *next;
#define BOUND_NODE_CHUNK_SIZE 256
   unsigned int num_taken;
   Bound_Node_Type nodes[BOUND_NODE_CHUNK_SIZE];
}
Bound_Node_Chunk_Type;

typedef struct
{
   SLang_Array_Type *at_x, *at_y;
   SLang_Array_Type *at_a_indices, *at_b_indices, *at_c_indices;
   Node_Type *tree;
   Node_Chunk_Type *node_chunks;
   Bound_Node_Chunk_Type *bound_node_chunks;
   Leaf_Node_Chunk_Type *leaf_node_chunks;
   double *xpts, *ypts;		       /* pointers into at_x, at_y */
   SLuindex_Type num_points;
   SLindex_Type *a_indices, *b_indices, *c_indices;   /* pointers into at_[abc]_indices */
   /* The nth triangle has vertices ABC given by
    *   A = (xpts[a_indices[nth]], ypts[a_indices[nth]]), etc.
    */
   SLuindex_Type num_triangles;
   SLindex_Type *indices;		       /* used by Node_Type */
}
Triangle_Search_Type;

#define MIN_NUM_LEAVES  9	       /* must be >= 2 */
#define BOUND_NODE_MIN_DEPTH 4
#define BOUND_NODE_DELTA_DEPTH 4

static Node_Type *alloc_node (Triangle_Search_Type *t)
{
   Node_Chunk_Type *chunk;
   Node_Type *node;

   if ((NULL == (chunk = t->node_chunks))
       || (chunk->num_taken == NODE_CHUNK_SIZE))
     {
	chunk = (Node_Chunk_Type *)SLmalloc (sizeof (Node_Chunk_Type));
	if (chunk == NULL)
	  return NULL;
	chunk->num_taken = 0;
	chunk->next = t->node_chunks;
	t->node_chunks = chunk;
     }
   node = chunk->nodes + chunk->num_taken;
   chunk->num_taken++;
   node->type = NODE_TYPE_NODE;
   return node;
}

static void dealloc_node (Node_Type *node)
{
   if (node == NULL)
     return;
   if (node->type == NODE_TYPE_BNDS_NODE)
     {
     }
   /* nothing to do */
}

static Bound_Node_Type *alloc_bound_node (Triangle_Search_Type *t)
{
   Bound_Node_Chunk_Type *chunk;
   Bound_Node_Type *node;

   if ((NULL == (chunk = t->bound_node_chunks))
       || (chunk->num_taken == BOUND_NODE_CHUNK_SIZE))
     {
	chunk = (Bound_Node_Chunk_Type *)SLmalloc (sizeof (Bound_Node_Chunk_Type));
	if (chunk == NULL)
	  return NULL;
	chunk->num_taken = 0;
	chunk->next = t->bound_node_chunks;
	t->bound_node_chunks = chunk;
     }
   node = chunk->nodes + chunk->num_taken;
   chunk->num_taken++;
   node->type = NODE_TYPE_BNDS_NODE;
   return node;
}

static void dealloc_bound_node (Bound_Node_Type *node)
{
   (void) node;
   /* nothing to do */
}

static Leaf_Node_Type *alloc_leaf_node (Triangle_Search_Type *t)
{
   Leaf_Node_Chunk_Type *chunk;
   Leaf_Node_Type *node;

   if ((NULL == (chunk = t->leaf_node_chunks))
       || (chunk->num_taken == LEAF_NODE_CHUNK_SIZE))
     {
	chunk = (Leaf_Node_Chunk_Type *)SLmalloc (sizeof (Leaf_Node_Chunk_Type));
	if (chunk == NULL)
	  return NULL;
	chunk->num_taken = 0;
	chunk->next = t->leaf_node_chunks;
	t->leaf_node_chunks = chunk;
     }
   node = chunk->nodes + chunk->num_taken;
   chunk->num_taken++;
   node->type = NODE_TYPE_LEAF;
   return node;
}

static void dealloc_leaf_node (Leaf_Node_Type *node)
{
   (void) node;
   /* nothing to do */
}

static void free_leaf_node_chunks (Leaf_Node_Chunk_Type *chunk)
{
   while (chunk != NULL)
     {
	Leaf_Node_Chunk_Type *next = chunk->next;
	SLfree ((char *) chunk);
	chunk = next;
     }
}

static void free_bound_node_chunks (Bound_Node_Chunk_Type *chunk)
{
   while (chunk != NULL)
     {
	Bound_Node_Chunk_Type *next = chunk->next;
	SLfree ((char *) chunk);
	chunk = next;
     }
}

static void free_node_chunks (Node_Chunk_Type *chunk)
{
   while (chunk != NULL)
     {
	Node_Chunk_Type *next = chunk->next;
	SLfree ((char *) chunk);
	chunk = next;
     }
}

static void free_tree (Node_Type *tree)
{
   if (tree == NULL)
     return;

   if (tree->type == NODE_TYPE_NODE)
     {
	free_tree (tree->left);
	free_tree (tree->right);
	dealloc_node (tree);
	return;
     }
   if (tree->type == NODE_TYPE_BNDS_NODE)
     {
	Bound_Node_Type *b = (Bound_Node_Type *)tree;
	free_tree (b->left);
	free_tree (b->right);
	dealloc_bound_node (b);
	return;
     }
   dealloc_leaf_node ((Leaf_Node_Type *)tree);
}

#define LOWEST_VERTEX_INDEX(x, i, j, k) \
   (((x[i]) < (x[j])) \
     ? (((x[i]) < (x[k])) ? (i) : (k)) \
     : ((x[j]) < (x[k]) ? (j) : (k)));

#define IS_BAD_VERTEX_INDEX(n,i) (((i)<0) || (((SLuindex_Type)i)>=n))
static void throw_vertex_error (SLuindex_Type maxindex)
{
   SLang_verror (SL_Index_Error, "Triangle vertex-index must be in the range 0<=index<%lu", (unsigned long)maxindex);
}



static double *Axis_Points;
static SLindex_Type *Lowest_X_Vertex_Indices, *Lowest_Y_Vertex_Indices;
static SLindex_Type *Lowest_Vertex_Indices;

static int sort_func (const void *ip, const void *jp)
{
   SLindex_Type i, j;
   double vi, vj;

   i = *(SLindex_Type *) ip;
   j = *(SLindex_Type *) jp;
   vi = Axis_Points[Lowest_Vertex_Indices[i]];
   vj = Axis_Points[Lowest_Vertex_Indices[j]];
   if (vi < vj)
     return -1;
   if (vi > vj)
     return 1;
   return 0;
}

static void minmax(double a, double b, double c,
		   double *xminp, double *xmaxp)
{
   double xmin = a, xmax = a;

   if (b < xmin) xmin = b; else if (b > xmax) xmax = b;
   if (c < xmin) xmin = c; else if (c > xmax) xmax = c;
   *xminp = xmin;
   *xmaxp = xmax;
}

static Leaf_Node_Type *create_leaf_node (Triangle_Search_Type *t,
					 SLindex_Type *indices, SLuindex_Type num_indices,
					 Bound_Type *bnds)
{
   Leaf_Node_Type *leaf;
   double xmin, ymin, xmax, ymax;
   SLindex_Type a, b, c;
   SLuindex_Type i;

   leaf = alloc_leaf_node (t);
   if (leaf == NULL)
     return NULL;

   i = indices[0];
   a = t->a_indices[i]; b = t->b_indices[i]; c = t->c_indices[i];
   minmax (t->xpts[a], t->xpts[b], t->xpts[c], &xmin, &xmax);
   minmax (t->ypts[a], t->ypts[b], t->ypts[c], &ymin, &ymax);

   for (i = 1; i < num_indices; i++)
     {
	double z0, z1;
	SLindex_Type j = indices[i];
	a = t->a_indices[j]; b = t->b_indices[j]; c = t->c_indices[j];
	minmax (t->xpts[a], t->xpts[b], t->xpts[c], &z0, &z1);
	if (z0 < xmin) xmin = z0;
	if (z1 > xmax) xmax = z1;
	minmax (t->ypts[a], t->ypts[b], t->ypts[c], &z0, &z1);
	if (z0 < ymin) ymin = z0;
	if (z1 > ymax) ymax = z1;
     }

   leaf->bnds.xmin = xmin;
   leaf->bnds.xmax = xmax;
   leaf->bnds.ymin = ymin;
   leaf->bnds.ymax = ymax;

   leaf->indices = indices;
   leaf->num_indices = num_indices;

   *bnds = leaf->bnds;
   return leaf;
}

static void merge_bounds (Bound_Type *a, Bound_Type *b, Bound_Type *ab)
{
   ab->xmin = (a->xmin < b->xmin) ? a->xmin : b->xmin;
   ab->xmax = (a->xmax > b->xmax) ? a->xmax : b->xmax;
   ab->ymin = (a->ymin < b->ymin) ? a->ymin : b->ymin;
   ab->ymax = (a->ymax > b->ymax) ? a->ymax : b->ymax;
}

static Node_Type *new_tree (Triangle_Search_Type *t,
			    SLindex_Type *indices,  SLuindex_Type num_indices,
			    Bound_Type *bnds,
			    int depth)
{
   Node_Type *node;
   SLuindex_Type num_left;
   SLuindex_Type i;
   Bound_Type lbnds, rbnds;

   if (num_indices <= MIN_NUM_LEAVES)
     return (Node_Type *)create_leaf_node (t, indices, num_indices, bnds);

   if ((depth >= BOUND_NODE_DELTA_DEPTH)
       && ((depth % BOUND_NODE_DELTA_DEPTH) == 0))
     node = (Node_Type*)alloc_bound_node (t);
   else
     node = alloc_node (t);

   if (node == NULL)
     return NULL;

   if (depth % 2)
     {
	Axis_Points = t->ypts;
	Lowest_Vertex_Indices = Lowest_Y_Vertex_Indices;
     }
   else
     {
	Axis_Points = t->xpts;
	Lowest_Vertex_Indices = Lowest_X_Vertex_Indices;
     }
   qsort((void *)indices, num_indices, sizeof(SLindex_Type), sort_func);

   depth++;
   num_left = num_indices/2;
   i = indices[num_left];
   node->median = Axis_Points[Lowest_Vertex_Indices[i]];

   if (NULL == (node->left = new_tree (t, indices, num_left, &lbnds, depth)))
     {
	dealloc_node (node);
	return NULL;
     }

   if (NULL == (node->right = new_tree (t, indices + num_left, num_indices-num_left, &rbnds, depth)))
     {
	free_tree (node->left);
	dealloc_node (node);
	return NULL;
     }
   merge_bounds (&lbnds, &rbnds, bnds);
   if (node->type == NODE_TYPE_BNDS_NODE)
     {
	Bound_Node_Type *b = (Bound_Node_Type *) node;
	b->bnds = *bnds;
     }
   return node;
}

static int create_search_tree (Triangle_Search_Type *t)
{
   SLindex_Type *indices;
   SLindex_Type *a_indices, *b_indices, *c_indices;
   SLuindex_Type i, num_indices, num_points;
   double *xpts, *ypts;
   SLindex_Type *indices_of_lowest_x_vertex;
   SLindex_Type *indices_of_lowest_y_vertex;
   Bound_Type bnds;
   int status;

   num_indices = t->num_triangles;
   indices = (SLindex_Type *) SLmalloc (sizeof(SLindex_Type)*(num_indices + 1));
   if (indices == NULL)
     return -1;
   indices_of_lowest_y_vertex = (SLindex_Type *)SLmalloc (sizeof(SLindex_Type)*(num_indices+1));
   if (indices_of_lowest_y_vertex == NULL)
     {
	SLfree ((char *) indices);
	return -1;
     }
   indices_of_lowest_x_vertex = (SLindex_Type *)SLmalloc (sizeof(SLindex_Type)*(num_indices+1));
   if (indices_of_lowest_x_vertex == NULL)
     {
	SLfree ((char *) indices_of_lowest_y_vertex);
	SLfree ((char *) indices);
	return -1;
     }

   xpts = t->xpts;
   ypts = t->ypts;
   num_points = t->num_points;

   status = 0;
   a_indices = t->a_indices;
   b_indices = t->b_indices;
   c_indices = t->c_indices;

   for (i = 0; i < num_indices; i++)
     {
	SLindex_Type a, b, c;
	a = a_indices[i];
	b = b_indices[i];
	c = c_indices[i];
	if (IS_BAD_VERTEX_INDEX(num_points, a)
	    || IS_BAD_VERTEX_INDEX(num_points, b)
	    || IS_BAD_VERTEX_INDEX(num_points, c))
	  {
	     throw_vertex_error (num_indices);
	     status = -1;
	     goto free_and_return;
	  }
	indices_of_lowest_x_vertex[i] = LOWEST_VERTEX_INDEX(xpts, a, b, c);
	indices_of_lowest_y_vertex[i] = LOWEST_VERTEX_INDEX(ypts, a, b, c);
	indices[i] = i;
     }

   t->indices = indices;
   Lowest_X_Vertex_Indices = indices_of_lowest_x_vertex;
   Lowest_Y_Vertex_Indices = indices_of_lowest_y_vertex;

   if (NULL == (t->tree = new_tree (t, indices, num_indices, &bnds, 0)))
     status = -1;

free_and_return:
   SLfree ((char *) indices_of_lowest_x_vertex);
   SLfree ((char *) indices_of_lowest_y_vertex);
   return status;
}


/* Determine if the point P(x,y) is in the triangle ABC.  Assume AB and AC is
 * a 2d basis:
 *
 *   P-A = alpha*(B-A) + beta*(C-A)
 *
 * If P is in/on the triangle, then 0 <= alpha+beta <= 1.  Obviously, we
 * need to solve the linear system:
 * 
 *   PA.BA = alpha*(BA.BA) + beta*(CA.BA)
 *   PA.CA = alpha*(BA.CA) + beta*(CA.CA)
 */
static int in_triangle (double x, double y,
			double xa, double xb, double xc,
			double ya, double yb, double yc,
			double *alphap, double *betap)
{
   double xba, yba, xca, yca, xpa, ypa, baba, baca, caca, paba, paca;
   double den, alpha, beta;
   double fba, fcb, fac;

   fba = (xb-xa)*(y-ya) - (x-xa)*(yb-ya);
   fcb = (xc-xb)*(y-yb) - (x-xb)*(yc-yb);
   fac = (xa-xc)*(y-yc) - (x-xc)*(ya-yc);
   if (fba > 0)
     {
	if ((fcb < 0) || (fac < 0))
	  return 0;
     }
   else if ((fba < 0) && ((fcb > 0) || (fac > 0)))
     return 0;

   xba = xb-xa; yba = yb-ya; xca = xc-xa; yca = yc-ya; xpa = x-xa; ypa = y-ya;

   baba = xba*xba + yba*yba;
   baca = xba*xca + yba*yca;
   caca = xca*xca + yca*yca;
   paba = xpa*xba + ypa*yba;
   paca = xpa*xca + ypa*yca;
   den = baba*caca - baca*baca;
   if (den == 0.0)
     return 0;
   alpha = (paba*caca - paca*baca)/den;
   beta = (baba*paca - paba*baca)/den;

   *alphap = alpha;
   *betap = beta;
   return 1;
}

static int find_triangle (Triangle_Search_Type *t, Node_Type *node, 
			  int depth, double x, double y,
			  SLindex_Type *idxp, double *alphap, double *betap)
{
   SLindex_Type *indices;
   SLuindex_Type i, num_indices, num_points;
   Leaf_Node_Type *leaf;

   if (node->type != NODE_TYPE_LEAF)
     {
	double val;

	if (node->type == NODE_TYPE_BNDS_NODE)
	  {
	     Bound_Node_Type *b = (Bound_Node_Type *)node;
	     if ((x < b->bnds.xmin) || (x > b->bnds.xmax)
		 || (y < b->bnds.ymin) || (y > b->bnds.ymax))
	       return 0;
	  }

	val = (depth % 2) ? y : x;

	depth++;
	if (val >= node->median)
	  {
	     if (1 == find_triangle (t, node->right, depth, x, y, idxp, alphap, betap))
	       return 1;
	  }
	if (1 == find_triangle (t, node->left, depth, x, y, idxp, alphap, betap))
	  return 1;

	return 0;
     }

   leaf = (Leaf_Node_Type *)node;
   if ((x < leaf->bnds.xmin) || (x > leaf->bnds.xmax)
       || (y < leaf->bnds.ymin) || (y > leaf->bnds.ymax))
     return 0;

   indices = leaf->indices;
   num_indices = leaf->num_indices;
   num_points = t->num_points;
   for (i = 0; i < num_indices; i++)
     {
	SLindex_Type idx, a, b, c;
	double xa, ya, xb, yb, xc, yc;

	idx = indices[i];
	a = t->a_indices[idx];
	if (IS_BAD_VERTEX_INDEX(num_points, a))
	  continue;
	xa = t->xpts[a];
	b = t->b_indices[idx];
	if (IS_BAD_VERTEX_INDEX(num_points, b))
	  continue;
	xb = t->xpts[b];
	c = t->c_indices[idx];
	if (IS_BAD_VERTEX_INDEX(num_points, c))
	  continue;
	xc = t->xpts[c];

	if (((x < xa) && (x < xb) && (x < xc))
	    || ((x > xa) && (x > xb) && (x > xc)))
	  continue;

	ya = t->ypts[a];
	yb = t->ypts[b];
	yc = t->ypts[c];

	if (((y < ya) && (y < yb) && (y < yc))
	    || ((y > ya) && (y > yb) && (y > yc)))
	  continue;

	if (1 == in_triangle (x, y, xa, xb, xc, ya, yb, yc, alphap, betap))
	  {
	     *idxp = idx;
	     return 1;
	  }
     }
   return 0;
}

static int find_triangle_containing_point (Triangle_Search_Type *t,
					   double x, double y,
					   SLindex_Type *idxp,
					   double *alphap, double *betap)
{
   return find_triangle (t, t->tree, 0, x, y, idxp, alphap, betap);
}

#define DUMMY_TRIANGLE_TYPE ((SLtype)-1)
static SLtype Triangle_Search_Type_Id = DUMMY_TRIANGLE_TYPE;

static int pop_2_double_arrays (SLang_Array_Type **atx, SLang_Array_Type **aty)
{
   SLang_Array_Type *x, *y;

   if (-1 == SLang_pop_array_of_type (&y, SLANG_DOUBLE_TYPE))
     return -1;
   if (-1 == SLang_pop_array_of_type (&x, SLANG_DOUBLE_TYPE))
     {
	SLang_free_array (y);
	return -1;
     }

   if (x->num_elements != y->num_elements)
     {
	SLang_free_array (x);
	SLang_free_array (y);
	SLang_verror (SL_InvalidParm_Error, "(X,Y) point arrays must have the same number of elements");
	return -1;
     }
   *atx = x;
   *aty = y;
   return 0;
}

static int pop_3_index_arrays (SLang_Array_Type **ap, SLang_Array_Type **bp, SLang_Array_Type **cp)
{
   SLang_Array_Type *a, *b, *c;

   if (-1 == SLang_pop_array_of_type (&c, SLANG_ARRAY_INDEX_TYPE))
     return -1;
   if (-1 == SLang_pop_array_of_type (&b, SLANG_ARRAY_INDEX_TYPE))
     {
	SLang_free_array (c);
	return -1;
     }
   if (-1 == SLang_pop_array_of_type (&a, SLANG_ARRAY_INDEX_TYPE))
     {
	SLang_free_array (b);
	SLang_free_array (c);
	return -1;
     }

   if ((a->num_elements != b->num_elements)
       || (a->num_elements != c->num_elements))
     {
	SLang_free_array (a);
	SLang_free_array (b);
	SLang_free_array (c);
	SLang_verror (SL_InvalidParm_Error, "Index vertex arrays must have the same number of elements");
	return -1;
     }
   *ap = a;
   *bp = b;
   *cp = c;
   return 0;
}

static void free_triangle_search_type (Triangle_Search_Type *t)
{
   if (t == NULL)
     return;
   SLang_free_array (t->at_a_indices);
   SLang_free_array (t->at_b_indices);
   SLang_free_array (t->at_c_indices);
   SLang_free_array (t->at_x);
   SLang_free_array (t->at_y);
   free_tree (t->tree);
   free_node_chunks (t->node_chunks);
   free_bound_node_chunks (t->bound_node_chunks);
   free_leaf_node_chunks (t->leaf_node_chunks);
   SLfree ((char *) t->indices);
   SLfree ((char *) t);
}

static int check_indices (SLindex_Type *indices, SLuindex_Type num, SLuindex_Type max_val)
{
   SLuindex_Type i;

   for (i = 0; i < num; i++)
     {
	if (IS_BAD_VERTEX_INDEX(max_val, indices[i]))
	  throw_vertex_error (max_val);
     }
   return 0;
}

/* t = open_triangle_search (xpts, ypts, a_indices, b_indices, c_indices) */
static void open_triangle_search_intrin (void)
{
   Triangle_Search_Type *t;
   SLang_MMT_Type *mmt;

   if (NULL == (t = (Triangle_Search_Type *)SLmalloc (sizeof (Triangle_Search_Type))))
     return;
   memset ((char *) t, 0, sizeof (Triangle_Search_Type));
   if (-1 == pop_3_index_arrays (&t->at_a_indices, &t->at_b_indices, &t->at_c_indices))
     {
	free_triangle_search_type (t);
	return;
     }
   if (-1 == pop_2_double_arrays (&t->at_x, &t->at_y))
     {
	free_triangle_search_type (t);
	return;
     }

   t->xpts = (double *)t->at_x->data;
   t->ypts = (double *)t->at_y->data;
   t->num_points = t->at_x->num_elements;
   t->a_indices = (SLindex_Type *)t->at_a_indices->data;
   t->b_indices = (SLindex_Type *)t->at_b_indices->data;
   t->c_indices = (SLindex_Type *)t->at_c_indices->data;
   t->num_triangles = t->at_a_indices->num_elements;

   if ((-1 == check_indices (t->a_indices, t->num_triangles, t->num_points))
       || (-1 == check_indices (t->b_indices, t->num_triangles, t->num_points))
       || (-1 == check_indices (t->c_indices, t->num_triangles, t->num_points)))
     {
	free_triangle_search_type (t);
	return;
     }
   if (-1 == create_search_tree (t))
     {
	free_triangle_search_type (t);
	return;
     }

   if (NULL == (mmt = SLang_create_mmt (Triangle_Search_Type_Id, (VOID_STAR) t)))
     {
	free_triangle_search_type (t);
	return;
     }

   if (-1 == SLang_push_mmt (mmt))
     {
	SLang_free_mmt (mmt);
	return;
     }
}

static void find_triangle_intrin (void)
{
   int is_scalar;
   double x, y, alpha, beta;
   double *xpts, *ypts, *alphap, *betap;
   SLindex_Type *indices;
   SLuindex_Type i, num_pts;
   SLang_Array_Type *atx, *aty, *at_idx, *at_alpha, *at_beta;
   SLang_Ref_Type *ralpha, *rbeta;
   SLang_MMT_Type *mmt;
   Triangle_Search_Type *t;
   int inc;

   at_alpha = at_beta = atx = aty = at_idx = NULL;
   ralpha = rbeta = NULL;
   mmt = NULL;

   if (SLang_Num_Function_Args == 5)
     {
	if (-1 == SLang_pop_ref (&rbeta))
	  return;
	if (-1 == SLang_pop_ref (&ralpha))
	  {
	     SLang_free_ref (rbeta);
	     return;
	  }
     }
   if (SLang_peek_at_stack () == SLANG_ARRAY_TYPE)
     {
	is_scalar = 0;
	if (-1 == pop_2_double_arrays (&atx, &aty))
	  goto free_and_return;
     }
   else
     {
	is_scalar = 1;
	if ((-1 == SLang_pop_double (&y))
	    || (-1 == SLang_pop_double (&x)))
	  goto free_and_return;
     }

   if (NULL == (mmt = SLang_pop_mmt (Triangle_Search_Type_Id)))
     goto free_and_return;
   t = (Triangle_Search_Type *) SLang_object_from_mmt (mmt);

   if (is_scalar)
     {
	SLindex_Type idx;
	if (1 != find_triangle_containing_point (t, x, y, &idx, &alpha, &beta))
	  idx = -1;
	(void) SLang_assign_to_ref (ralpha, SLANG_DOUBLE_TYPE, &alpha);
	(void) SLang_assign_to_ref (rbeta, SLANG_DOUBLE_TYPE, &beta);
	(void) SLang_push_int (idx);
	goto free_and_return;
     }

   at_idx = SLang_create_array1 (SLANG_INT_TYPE, 0, NULL, atx->dims, atx->num_dims, 0);
   if (at_idx == NULL)
     goto free_and_return;

   inc = 0;
   alphap = &alpha;
   betap = &beta;
   if (ralpha != NULL)
     {
	if (NULL == (at_alpha = SLang_create_array1 (SLANG_DOUBLE_TYPE, 0, NULL, atx->dims, atx->num_dims, 0))
	    || (NULL == (at_beta = SLang_create_array1 (SLANG_DOUBLE_TYPE, 0, NULL, atx->dims, atx->num_dims, 0))))
	  goto free_and_return;
	inc = 1;
	alphap = (double *)at_alpha->data;
	betap = (double *)at_beta->data;
     }

   num_pts = atx->num_elements;
   xpts = (double *) atx->data;
   ypts = (double *) aty->data;
   indices = (SLindex_Type *) at_idx->data;
   for (i = 0; i < num_pts; i++)
     {
	SLindex_Type idx;
	if (1 != find_triangle_containing_point (t, xpts[i], ypts[i], &idx, alphap, betap))
	  {
	     idx = -1;
	     *alphap = -1;
	     *betap = -1;
	  }
	indices[i] = idx;
	alphap += inc;
	betap += inc;
     }
   (void) SLang_push_array (at_idx, 0);
   if (ralpha != NULL)
     {
	(void) SLang_assign_to_ref (ralpha, SLANG_ARRAY_TYPE, &at_alpha);
	(void) SLang_assign_to_ref (rbeta, SLANG_ARRAY_TYPE, &at_beta);
     }
   /* drop */
free_and_return:
   SLang_free_ref (ralpha);
   SLang_free_ref (rbeta);
   SLang_free_mmt (mmt);
   SLang_free_array (at_alpha);
   SLang_free_array (at_beta);
   SLang_free_array (atx);
   SLang_free_array (aty);
   SLang_free_array (at_idx);
}

static double compute_area (double xa, double xb, double xc,
			    double ya, double yb, double yc)
{
   double xba, yba, xca, yca;
   xba = xb-xa; yba = yb-ya; xca = xc-xa; yca = yc-ya;
   return fabs (xba*yca - yba*xca);
}

/* areas = triangle_area (xpts, ypts, v0, v1, v2) */
static void triangle_area_intrin (void)
{
   int is_scalar;
   double *xpts, *ypts, *areas;
   SLuindex_Type i, num_points, num_triangles;
   SLang_Array_Type *atx = NULL, *aty = NULL;
   SLang_Array_Type *atv0 = NULL, *atv1 = NULL, *atv2 = NULL;
   SLindex_Type v0, v1, v2;
   SLindex_Type *v0p, *v1p, *v2p;
   SLang_Array_Type *at_area = NULL;
   double area;

   if (SLang_peek_at_stack () == SLANG_ARRAY_TYPE)
     {
	if (-1 == pop_3_index_arrays (&atv0, &atv1, &atv2))
	  return;
	is_scalar = 0;
     }
   else
     {
	if ((-1 == SLang_pop_array_index (&v2))
	    || (-1 == SLang_pop_array_index (&v1))
	    || (-1 == SLang_pop_array_index (&v0)))
	  return;
	is_scalar = 1;
     }

   if (-1 == pop_2_double_arrays (&atx, &aty))
     goto free_and_return;

   num_points = atx->num_elements;
   xpts = (double *) atx->data;
   ypts = (double *) aty->data;

   if (is_scalar)
     {
	if (IS_BAD_VERTEX_INDEX(num_points, v0)
	    || IS_BAD_VERTEX_INDEX(num_points, v1)
	    || IS_BAD_VERTEX_INDEX(num_points, v2))
	  {
	     throw_vertex_error (num_points);
	     goto scalar_free_and_return;
	  }
	area = compute_area (xpts[v0], xpts[v1], xpts[v2],
			     ypts[v0], ypts[v1], ypts[v2]);
	(void) SLang_push_double (area);
	goto scalar_free_and_return;
     }

   v0p = (SLindex_Type *) atv0->data;
   v1p = (SLindex_Type *) atv1->data;
   v2p = (SLindex_Type *) atv2->data;
   num_triangles = atv0->num_elements;

   at_area = SLang_create_array1 (SLANG_DOUBLE_TYPE, 0, NULL, atv0->dims, atv0->num_dims, 0);
   if (at_area == NULL)
     goto free_and_return;
   areas = (double *)at_area->data;

   for (i = 0; i < num_triangles; i++)
     {
	v0 = v0p[i]; v1 = v1p[i]; v2 = v2p[i];
	if (IS_BAD_VERTEX_INDEX(num_points, v0)
	    || IS_BAD_VERTEX_INDEX(num_points, v1)
	    || IS_BAD_VERTEX_INDEX(num_points, v2))
	  {
	     throw_vertex_error (num_points);
	     goto free_and_return;
	  }
	areas[i] = compute_area (xpts[v0], xpts[v1], xpts[v2],
				 ypts[v0], ypts[v1], ypts[v2]);
     }
   (void) SLang_push_array (at_area, 0);

   /* drop */

free_and_return:
   SLang_free_array (atv0);
   SLang_free_array (atv1);
   SLang_free_array (atv2);
   SLang_free_array (at_area);
scalar_free_and_return:
   SLang_free_array (atx);
   SLang_free_array (aty);
}

static void triangulate_intrin (void)
{
   SLang_Array_Type *at_x, *at_y;
   Voronoi_Triangulate_Type tri;

   if (-1 == pop_2_double_arrays (&at_x, &at_y))
     return;

   if (0 == voronoi_triangulate ((double *)at_x->data, (double *)at_y->data,
				 at_x->num_elements, &tri))
     {
	unsigned int i;
	SLindex_Type dims = (SLindex_Type) tri.num_triangles;
	for (i = 0; i < 3; i++)
	  {
	     SLang_Array_Type *v;
	     v = SLang_create_array (SLANG_INT_TYPE, 0, tri.vertex_indices[i], &dims, 1);
	     if (v != NULL)
	       {
		  if (0 == SLang_push_array (v, 1))
		    continue;
		  /* Otherwise, push failed and v has been freed */
		  i++;
		  /* drop */
	       }

	     while (i < 3)
	       {
		  SLfree ((char *)tri.vertex_indices[i]);
		  i++;
	       }
	  }
     }
   SLang_free_array (at_y);
   SLang_free_array (at_x);
}

static void convex_hull_intrin (void)
{
   SLang_Array_Type *at_x, *at_y;
   int *indices;
   unsigned int num_indices;

   if (-1 == pop_2_double_arrays (&at_x, &at_y))
     return;

   if (0 == voronoi_convex_hull ((double *)at_x->data, (double *)at_y->data,
				 at_x->num_elements, &indices, &num_indices))
     {
	SLang_Array_Type *at;
	SLindex_Type dims;

	dims = num_indices;
	at = SLang_create_array (SLANG_INT_TYPE, 0, indices, &dims, 1);
	if (at != NULL)
	  (void) SLang_push_array (at, 1);
	/* drop */
     }
   SLang_free_array (at_y);
   SLang_free_array (at_x);
}

static SLang_Intrin_Fun_Type Module_Intrinsics [] =
{
   MAKE_INTRINSIC_0("_voronoi_triangulate", triangulate_intrin, SLANG_VOID_TYPE),
   MAKE_INTRINSIC_0("_voronoi_triangle_new_search", open_triangle_search_intrin, SLANG_VOID_TYPE),
   MAKE_INTRINSIC_0("_voronoi_triangle_search", find_triangle_intrin, SLANG_VOID_TYPE),
   MAKE_INTRINSIC_0("_voronoi_triangle_area", triangle_area_intrin, SLANG_VOID_TYPE),
   MAKE_INTRINSIC_0("_voronoi_convex_hull", convex_hull_intrin, SLANG_VOID_TYPE),
   SLANG_END_INTRIN_FUN_TABLE
};

static char *Module_Version_String = MODULE_VERSION_STRING;
static SLang_Intrin_Var_Type Module_Variables [] =
{
   MAKE_VARIABLE("_voronoi_module_version_string", &Module_Version_String, SLANG_STRING_TYPE, 1),
   SLANG_END_INTRIN_VAR_TABLE
};

static SLang_IConstant_Type Module_IConstants [] =
{
   MAKE_ICONSTANT("_voronoi_module_version", MODULE_VERSION_NUMBER),
   SLANG_END_ICONST_TABLE
};

static SLang_DConstant_Type Module_DConstants [] =
{
   SLANG_END_DCONST_TABLE
};

static void destroy_triangle_search_type (SLtype type, VOID_STAR ptr)
{
   (void) type;
   free_triangle_search_type ((Triangle_Search_Type *) ptr);
}

static int register_module_classes (void)
{
   SLang_Class_Type *cl;

   if (Triangle_Search_Type_Id != DUMMY_TRIANGLE_TYPE)
     return 0;

   if (NULL == (cl = SLclass_allocate_class ("Triangle_Search_Type")))
     return -1;

   if (-1 == SLclass_set_destroy_function (cl, destroy_triangle_search_type))
     return -1;

   if (-1 == SLclass_register_class (cl, SLANG_VOID_TYPE, sizeof (Triangle_Search_Type), SLANG_CLASS_TYPE_MMT))
     return -1;

   Triangle_Search_Type_Id = SLclass_get_class_id (cl);
   if (-1 == SLclass_patch_intrin_fun_table1 (Module_Intrinsics, DUMMY_TRIANGLE_TYPE, Triangle_Search_Type_Id))
     return -1;

   return 0;
}

int init_voronoi_module_ns (char *ns_name)
{
   SLang_NameSpace_Type *ns = SLns_create_namespace (ns_name);
   if (ns == NULL)
     return -1;

   if ((-1 == register_module_classes ())
       || (-1 == SLns_add_intrin_var_table (ns, Module_Variables, NULL))
       || (-1 == SLns_add_intrin_fun_table (ns, Module_Intrinsics, NULL))
       || (-1 == SLns_add_iconstant_table (ns, Module_IConstants, NULL))
       || (-1 == SLns_add_dconstant_table (ns, Module_DConstants, NULL))
       )
     return -1;

   return 0;
}

/* This function is optional */
void deinit_voronoi_module (void)
{
}
